use sprinklers_core::zone_interface::{ZoneInterface, ZoneNum};

use eyre::WrapErr;
use gpio_cdev::{LineHandle, LineRequestFlags};
use serde::{Deserialize, Serialize};
use tracing::{error, trace, warn};

pub struct LinuxGpio {
    lines: Vec<LineHandle>,
}

impl ZoneInterface for LinuxGpio {
    fn num_zones(&self) -> ZoneNum {
        self.lines.len() as ZoneNum
    }

    fn set_zone_state(&self, id: ZoneNum, running: bool) {
        if let Some(line) = &self.lines.get(id as usize) {
            trace!(
                line = line.line().offset(),
                id,
                running,
                "setting state of line"
            );
            if let Err(err) = line.set_value(running as u8) {
                error!("error setting GPIO line value: {}", err);
            }
        } else {
            warn!("set_zone_state: invalid zone id: {}", id);
        }
    }

    fn get_zone_state(&self, id: ZoneNum) -> bool {
        if let Some(line) = &self.lines.get(id as usize) {
            match line.get_value() {
                Ok(active) => active != 0,
                Err(err) => {
                    error!("error getting GPIO line value: {}", err);
                    false
                }
            }
        } else {
            warn!("get_zone_state: invalid zone id: {}", id);
            false
        }
    }
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct LinuxGpioConfig {
    chip_path: String,
    line_offsets: Vec<u32>,
}

impl LinuxGpioConfig {
    pub fn build(self) -> eyre::Result<LinuxGpio> {
        let mut chip =
            gpio_cdev::Chip::new(self.chip_path).wrap_err("could not create gpio_cdev Chip")?;
        let lines: Result<Vec<_>, eyre::Report> = self
            .line_offsets
            .into_iter()
            .map(|line_offset| {
                let line = chip
                    .get_line(line_offset)
                    .wrap_err("could not get line for chip")?;
                let line_handle = line
                    .request(LineRequestFlags::OUTPUT, 0, "sprinklers_rs")
                    .wrap_err("could not request line access")?;
                Ok(line_handle)
            })
            .collect();
        let lines = lines?;
        Ok(LinuxGpio { lines })
    }
}
