use super::*;
use sprinklers_actors::zone_runner::ZoneRunHandle;
use sprinklers_core::model::{self, ZoneRef};
use sprinklers_core::serde::duration_secs;

use eyre::WrapErr;
use serde::{Deserialize, Serialize};
use std::time::Duration;

#[derive(Copy, Clone, Debug, Serialize, Deserialize)]
#[serde(transparent)]
pub struct ZoneId(pub model::ZoneId);

impl ZoneId {
    fn get_zone(self, zones: &Zones) -> Result<ZoneRef, RequestError> {
        zones
            .get(&self.0)
            .cloned()
            .ok_or_else(|| RequestError::with_name(ErrorCode::NoSuchZone, "no such zone", "zone"))
    }
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct RunZoneRequest {
    // TODO: update nomenclature
    #[serde(rename = "sectionId")]
    pub zone_id: ZoneId,
    #[serde(with = "duration_secs")]
    pub duration: Duration,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct RunZoneResponse {
    pub message: String,
    pub run_id: ZoneRunHandle,
}

impl IRequest for RunZoneRequest {
    type Response = RunZoneResponse;
    fn exec(self, ctx: &mut RequestContext) -> RequestFuture<Self::Response> {
        let mut zone_runner = ctx.zone_runner.clone();
        let zone = self.zone_id.get_zone(&*ctx.zones.borrow());
        let duration = self.duration;
        Box::pin(async move {
            let zone = zone?;
            let handle = zone_runner
                .queue_run(zone.clone(), duration)
                .await
                .wrap_err("could not queue run")?;
            Ok(RunZoneResponse {
                message: format!("running zone '{}' for {:?}", &zone.name, duration),
                run_id: handle,
            })
        })
    }
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct CancelZoneRequest {
    // TODO: update nomenclature
    #[serde(rename = "sectionId")]
    pub zone_id: ZoneId,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct CancelZoneResponse {
    pub message: String,
    pub cancelled: usize,
}

impl IRequest for CancelZoneRequest {
    type Response = CancelZoneResponse;
    fn exec(self, ctx: &mut RequestContext) -> RequestFuture<Self::Response> {
        let mut zone_runner = ctx.zone_runner.clone();
        let zone = self.zone_id.get_zone(&*ctx.zones.borrow());
        Box::pin(async move {
            let zone = zone?;
            let cancelled = zone_runner
                .cancel_by_zone(zone.id)
                .await
                .wrap_err("could not cancel zone")?;
            Ok(CancelZoneResponse {
                message: format!("cancelled {} runs for zone '{}'", cancelled, zone.name),
                cancelled,
            })
        })
    }
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct CancelZoneRunIdRequest {
    pub run_id: ZoneRunHandle,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct CancelZoneRunIdResponse {
    pub message: String,
    pub cancelled: bool,
}

impl IRequest for CancelZoneRunIdRequest {
    type Response = ResponseMessage;
    fn exec(self, ctx: &mut RequestContext) -> RequestFuture<Self::Response> {
        let mut zone_runner = ctx.zone_runner.clone();
        Box::pin(async move {
            let cancelled = zone_runner
                .cancel_run(self.run_id)
                .await
                .wrap_err("could not cancel zone run")?;
            if cancelled {
                Ok(ResponseMessage::new("cancelled zone run"))
            } else {
                Err(RequestError::with_name(
                    ErrorCode::NoSuchZoneRun,
                    "no such zone run",
                    "zone run",
                ))
            }
        })
    }
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct PauseZoneRunnerRequest {
    pub paused: bool,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct PauseZoneRunnerResponse {
    pub message: String,
    pub paused: bool,
}

impl IRequest for PauseZoneRunnerRequest {
    type Response = PauseZoneRunnerResponse;
    fn exec(self, ctx: &mut RequestContext) -> RequestFuture<Self::Response> {
        let mut zone_runner = ctx.zone_runner.clone();
        let paused = self.paused;
        Box::pin(async move {
            if paused {
                zone_runner.pause().await
            } else {
                zone_runner.unpause().await
            }
            .wrap_err("could not pause/unpause zone runner")?;
            Ok(PauseZoneRunnerResponse {
                message: format!("{} zone runner", if paused { "paused" } else { "unpaused" }),
                paused,
            })
        })
    }
}
