use super::MqttInterface;
use sprinklers_actors::{
    program_runner::{ProgramEvent, ProgramEventRecv},
    zone_runner::{ZoneEvent, ZoneEventRecv, ZoneRunnerState, ZoneRunnerStateRecv},
};

use actix::{fut::wrap_future, Actor, ActorContext, Addr, AsyncContext, Handler, StreamHandler};
use futures_util::TryFutureExt;
use sprinklers_core::model::{Programs, Zones};
use tokio::sync::{broadcast, watch};
use tracing::{trace, warn};

struct UpdateListenerActor {
    mqtt_interface: MqttInterface,
    old_zones: Option<Zones>,
    old_programs: Option<Programs>,
}

impl UpdateListenerActor {
    fn new(mqtt_interface: MqttInterface) -> Self {
        Self {
            mqtt_interface,
            old_zones: None,
            old_programs: None,
        }
    }
}

impl Actor for UpdateListenerActor {
    type Context = actix::Context<Self>;

    fn started(&mut self, _ctx: &mut Self::Context) {
        trace!("starting UpdateListener");
    }

    fn stopped(&mut self, _ctx: &mut Self::Context) {
        trace!("stopped UpdateListener")
    }
}

impl StreamHandler<Zones> for UpdateListenerActor {
    fn handle(&mut self, zones: Zones, ctx: &mut Self::Context) {
        let mut mqtt_interface = self.mqtt_interface.clone();

        let old_zones = self.old_zones.replace(zones.clone());

        let fut = async move {
            mqtt_interface.publish_zones(&zones).await?;
            for zone_id in zones.keys() {
                mqtt_interface.publish_zone_state(*zone_id, false).await?;
            }

            match old_zones {
                None => {
                    mqtt_interface.publish_zones(&zones).await?;

                    // Some what of a hack
                    // Initialize zone running states to false the first time we
                    // receive zones
                    for zone_id in zones.keys() {
                        mqtt_interface.publish_zone_state(*zone_id, false).await?;
                    }
                }
                Some(old_zones) => {
                    mqtt_interface
                        .publish_zones_diff(&old_zones, &zones)
                        .await?;
                }
            }
            Ok(())
        }
        .unwrap_or_else(|err: eyre::Report| warn!("could not publish programs: {:?}", err));
        ctx.spawn(wrap_future(fut));
    }
}

impl StreamHandler<Result<ZoneEvent, broadcast::RecvError>> for UpdateListenerActor {
    fn handle(&mut self, event: Result<ZoneEvent, broadcast::RecvError>, ctx: &mut Self::Context) {
        let event = match event {
            Ok(ev) => ev,
            Err(broadcast::RecvError::Closed) => unreachable!(),
            Err(broadcast::RecvError::Lagged(n)) => {
                warn!("zone events lagged by {}", n);
                return;
            }
        };
        if let Some((zone_id, state)) = match event {
            ZoneEvent::RunStart(_, zone) | ZoneEvent::RunUnpause(_, zone) => Some((zone.id, true)),
            ZoneEvent::RunFinish(_, zone)
            | ZoneEvent::RunPause(_, zone)
            | ZoneEvent::RunCancel(_, zone) => Some((zone.id, false)),
            ZoneEvent::RunnerPause | ZoneEvent::RunnerUnpause => None,
        } {
            let mut mqtt_interface = self.mqtt_interface.clone();
            let fut = async move {
                if let Err(err) = mqtt_interface.publish_zone_state(zone_id, state).await {
                    warn!("could not publish zone state: {}", err);
                }
            };
            ctx.spawn(wrap_future(fut));
        }
    }
}

impl StreamHandler<Result<ProgramEvent, broadcast::RecvError>> for UpdateListenerActor {
    fn handle(
        &mut self,
        event: Result<ProgramEvent, broadcast::RecvError>,
        ctx: &mut Self::Context,
    ) {
        let event = match event {
            Ok(ev) => ev,
            Err(broadcast::RecvError::Closed) => unreachable!(),
            Err(broadcast::RecvError::Lagged(n)) => {
                warn!("program events lagged by {}", n);
                return;
            }
        };
        let mut mqtt_interface = self.mqtt_interface.clone();
        let fut = async move {
            enum Publish {
                Running(bool),
                NextRun(chrono::DateTime<chrono::Local>),
            }

            let (program_id, publish) = match event {
                ProgramEvent::RunStart(prog) => (prog.id, Publish::Running(true)),
                ProgramEvent::RunFinish(prog) | ProgramEvent::RunCancel(prog) => {
                    (prog.id, Publish::Running(false))
                }
                ProgramEvent::NextRun(prog, next_run) => (prog.id, Publish::NextRun(next_run)),
            };
            match publish {
                Publish::Running(running) => {
                    if let Err(err) = mqtt_interface
                        .publish_program_running(program_id, running)
                        .await
                    {
                        warn!("could not publish program running: {}", err);
                    }
                }
                Publish::NextRun(next_run) => {
                    if let Err(err) = mqtt_interface
                        .publish_program_next_run(program_id, next_run)
                        .await
                    {
                        warn!("could not publish program next run: {}", err);
                    }
                }
            }
        };
        ctx.spawn(wrap_future(fut));
    }
}

impl StreamHandler<ZoneRunnerState> for UpdateListenerActor {
    fn handle(&mut self, state: ZoneRunnerState, ctx: &mut Self::Context) {
        let mut mqtt_interface = self.mqtt_interface.clone();
        let fut = async move {
            if let Err(err) = mqtt_interface.publish_zone_runner(&state).await {
                warn!("could not publish zone runner: {}", err);
            }
        };
        ctx.spawn(wrap_future(fut));
    }
}

impl StreamHandler<Programs> for UpdateListenerActor {
    fn handle(&mut self, programs: Programs, ctx: &mut Self::Context) {
        let mut mqtt_interface = self.mqtt_interface.clone();

        let old_programs = self.old_programs.replace(programs.clone());

        let fut = async move {
            match old_programs {
                None => {
                    mqtt_interface.publish_programs(&programs).await?;

                    // Some what of a hack
                    // Initialize program running states to false the first time we
                    // receive programs
                    for program_id in programs.keys() {
                        mqtt_interface
                            .publish_program_running(*program_id, false)
                            .await?;
                    }
                }
                Some(old_programs) => {
                    mqtt_interface
                        .publish_programs_diff(&old_programs, &programs)
                        .await?;
                }
            }
            Ok(())
        }
        .unwrap_or_else(|err: eyre::Report| warn!("could not publish programs: {:?}", err));
        ctx.spawn(wrap_future(fut));
    }
}

#[derive(actix::Message)]
#[rtype(result = "()")]
struct Quit;

impl Handler<Quit> for UpdateListenerActor {
    type Result = ();

    fn handle(&mut self, _msg: Quit, ctx: &mut Self::Context) -> Self::Result {
        ctx.stop();
    }
}

trait Listenable<A>: Send
where
    A: Actor,
{
    fn listen(self, ctx: &mut A::Context);
}

#[derive(actix::Message)]
#[rtype(result = "()")]
struct Listen<L>(L)
where
    L: Listenable<UpdateListenerActor>;

impl<L> Handler<Listen<L>> for UpdateListenerActor
where
    L: Listenable<Self>,
{
    type Result = ();

    fn handle(&mut self, msg: Listen<L>, ctx: &mut Self::Context) -> Self::Result {
        msg.0.listen(ctx);
    }
}

impl Listenable<UpdateListenerActor> for watch::Receiver<Zones> {
    fn listen(self, ctx: &mut <UpdateListenerActor as Actor>::Context) {
        ctx.add_stream(self);
    }
}

impl Listenable<UpdateListenerActor> for ZoneEventRecv {
    fn listen(self, ctx: &mut <UpdateListenerActor as Actor>::Context) {
        ctx.add_stream(self);
    }
}

impl Listenable<UpdateListenerActor> for ZoneRunnerStateRecv {
    fn listen(self, ctx: &mut <UpdateListenerActor as Actor>::Context) {
        ctx.add_stream(self);
    }
}

impl Listenable<UpdateListenerActor> for watch::Receiver<Programs> {
    fn listen(self, ctx: &mut <UpdateListenerActor as Actor>::Context) {
        ctx.add_stream(self);
    }
}

impl Listenable<UpdateListenerActor> for ProgramEventRecv {
    fn listen(self, ctx: &mut <UpdateListenerActor as Actor>::Context) {
        ctx.add_stream(self);
    }
}

pub struct UpdateListener {
    addr: Addr<UpdateListenerActor>,
}

impl UpdateListener {
    pub fn start(mqtt_interface: MqttInterface) -> Self {
        let addr = UpdateListenerActor::new(mqtt_interface).start();
        Self { addr }
    }

    fn listen<L: 'static>(&mut self, listener: L)
    where
        L: Listenable<UpdateListenerActor>,
    {
        self.addr.do_send(Listen(listener));
    }

    pub fn listen_zones(&mut self, zones: watch::Receiver<Zones>) {
        self.listen(zones);
    }

    pub fn listen_zone_events(&mut self, zone_events: ZoneEventRecv) {
        self.listen(zone_events);
    }

    pub fn listen_zone_runner(&mut self, zone_runner_state_recv: ZoneRunnerStateRecv) {
        self.listen(zone_runner_state_recv);
    }

    pub fn listen_programs(&mut self, programs: watch::Receiver<Programs>) {
        self.listen(programs);
    }

    pub fn listen_program_events(&mut self, program_events: ProgramEventRecv) {
        self.listen(program_events);
    }

    pub async fn quit(self) -> eyre::Result<()> {
        Ok(self.addr.send(Quit).await?)
    }
}
