use sprinklers_actors::zone_runner::{ZoneRun, ZoneRunState, ZoneRunnerState};
use sprinklers_core::model::ZoneId;

use chrono::{DateTime, Utc};
use serde::Serialize;
use std::time::SystemTime;
use tokio::time::Instant;

#[derive(Clone, Debug, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct ZoneRunJson {
    id: i32,
    // TODO: change nomenclature
    #[serde(rename = "section")]
    zone: ZoneId,
    total_duration: f64,
    duration: f64,
    start_time: Option<String>,
    pause_time: Option<String>,
    unpause_time: Option<String>,
}

impl ZoneRunJson {
    fn from_run(run: &ZoneRun) -> Option<Self> {
        let (now, system_now) = (Instant::now(), SystemTime::now());
        let instant_to_string = |instant: Instant| -> String {
            DateTime::<Utc>::from(system_now - now.duration_since(instant)).to_rfc3339()
        };
        let (start_time, pause_time) = match run.state {
            ZoneRunState::Finished | ZoneRunState::Cancelled => {
                return None;
            }
            ZoneRunState::Waiting => (None, None),
            ZoneRunState::Running { start_time } => (Some(instant_to_string(start_time)), None),
            ZoneRunState::Paused {
                start_time,
                pause_time,
            } => (
                Some(instant_to_string(start_time)),
                Some(instant_to_string(pause_time)),
            ),
        };
        Some(Self {
            id: run.handle.clone().into_inner(),
            zone: run.zone.id,
            total_duration: run.total_duration.as_secs_f64(),
            duration: run.duration.as_secs_f64(),
            start_time,
            pause_time,
            unpause_time: None, // TODO: this is kinda useless, should probably be removed
        })
    }
}

#[derive(Clone, Debug, Serialize)]
#[serde(rename_all = "camelCase")]
pub struct ZoneRunnerStateJson {
    queue: Vec<ZoneRunJson>,
    current: Option<ZoneRunJson>,
    paused: bool,
}

impl From<&ZoneRunnerState> for ZoneRunnerStateJson {
    fn from(state: &ZoneRunnerState) -> Self {
        let mut run_queue = state.run_queue.iter();
        let current = run_queue.next().and_then(|run| ZoneRunJson::from_run(run));
        let queue = run_queue
            .filter_map(|run| ZoneRunJson::from_run(run))
            .collect();
        Self {
            queue,
            current,
            paused: state.paused,
        }
    }
}
