#![warn(clippy::all)]
#![warn(clippy::print_stdout)]

// mod option_future;
mod settings;
mod state_manager;
mod zone_interface;

use sprinklers_actors as actors;
use sprinklers_database as database;
use sprinklers_mqtt as mqtt;

use eyre::{Result, WrapErr};
use settings::Settings;
use tracing::info;
use tracing_subscriber::EnvFilter;

#[actix_rt::main]
async fn main() -> Result<()> {
    color_eyre::install()?;
    tracing_subscriber::fmt()
        .with_ansi(true)
        .with_env_filter(
            EnvFilter::try_from_default_env().unwrap_or_else(|_| EnvFilter::new("info")),
        )
        .init();

    let settings: Settings = Settings::new().wrap_err("could not load settings")?;

    info!("Starting sprinklers_rs...");

    let db_conn = database::setup_db()?;

    let zone_interface = settings.zone_interface.build()?;
    let mut zone_runner = actors::ZoneRunner::new(zone_interface);
    let mut program_runner = actors::ProgramRunner::new(zone_runner.clone());

    let state_manager = crate::state_manager::StateManagerThread::start(db_conn);

    let mqtt_options = settings.mqtt;
    // TODO: have ability to update zones / other data
    let request_context = mqtt::RequestContext {
        zones: state_manager.get_zones(),
        zone_runner: zone_runner.clone(),
        program_runner: program_runner.clone(),
        state_manager: state_manager.clone(),
    };
    let mqtt_interface = mqtt::MqttInterfaceTask::start(mqtt_options, request_context);

    let mut update_listener = mqtt::UpdateListener::start(mqtt_interface.clone());
    update_listener.listen_zones(state_manager.get_zones());
    update_listener.listen_zone_events(zone_runner.subscribe().await?);
    update_listener.listen_zone_runner(zone_runner.get_state_recv());
    update_listener.listen_programs(state_manager.get_programs());
    update_listener.listen_program_events(program_runner.subscribe().await?);

    // Only listen to programs now so above subscriptions get events
    program_runner.listen_zones(state_manager.get_zones());
    program_runner.listen_programs(state_manager.get_programs());

    info!("sprinklers_rs initialized");

    tokio::signal::ctrl_c().await?;
    info!("Interrupt received, shutting down");

    update_listener.quit().await?;
    mqtt_interface.quit().await?;
    drop(state_manager);
    program_runner.quit().await?;
    zone_runner.quit().await?;
    actix::System::current().stop();

    Ok(())
}
